@@ -33,6 +33,7 @@
 #endif /* HAVE_ERROR_H */
 #include <ctype.h>
 #include <errno.h>
+#include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -58,6 +59,26 @@
 #define PATCH "patch"
 #endif
 
+/* Line type for coloring */
+enum line_type {
+	LINE_FILE,
+	LINE_HEADER,
+	LINE_HUNK,
+	LINE_ADDED,
+	LINE_REMOVED,
+	LINE_TYPE_MAX
+};
+
+/* ANSI color codes for diff output */
+static const char *const color_codes[LINE_TYPE_MAX] = {
+
+	[LINE_FILE] = "\033[1m",      /* Bold for filenames */
+	[LINE_HEADER] = "\033[1m",    /* Bold for headers */
+	[LINE_HUNK] = "\033[36m",     /* Cyan for hunk headers */
+	[LINE_ADDED] = "\033[32m",    /* Green for added lines */
+	[LINE_REMOVED] = "\033[31m"   /* Red for removed lines */
+};
+
 /* This can be invoked as interdiff, combinediff, or flipdiff. */
 static enum {
 	mode_inter,
@@ -98,6 +119,8 @@ static int ignore_components = 0;
 static int ignore_components_specified = 0;
 static int unzip = 0;
 static int no_revert_omitted = 0;
+static int use_colors = 0;
+static int color_option_specified = 0;
 static int debug = 0;
 
 static struct patlist *pat_drop_context = NULL;
@@ -106,15 +129,42 @@ static struct file_list *files_done = NULL;
 static struct file_list *files_in_patch2 = NULL;
 static struct file_list *files_in_patch1 = NULL;
 
-/* checks whether file needs processing and sets context */
+/*
+ * Print colored output using a variadic format string.
+ */
+static void __attribute__((__format__(printf, 3, 4)))
+print_color (FILE *output_file, enum line_type type, const char *format, ...)
+{
+	const char *color_start = NULL;
+	va_list args;
+
+	/* Only colorize if colors are enabled AND we're outputting to stdout */
+	if (use_colors && output_file == stdout)
+		color_start = color_codes[type];
+
+	/* Print color start code */
+	if (color_start)
+		fputs (color_start, output_file);
+
+	/* Print the formatted content */
+	va_start (args, format);
+	vfprintf (output_file, format, args);
+	va_end (args);
+
+	/* Print color end code */
+	if (color_start)
+		fputs ("\033[0m", output_file);
+}
+
+ /* checks whether file needs processing and sets context */
 static int
 check_filename (const char *fn)
 {
-	if (patlist_match(pat_drop_context, fn)) {
+	if (patlist_match(pat_drop_context, fn))
 		max_context = 0;
-	} else {
+	else
 		max_context = max_context_real;
-	}
+
 	return 1;
 }
 
@@ -181,9 +231,8 @@ determine_ignore_components (struct file_list *list1, struct file_list *list2)
 			const char *stripped1 = stripped(l1->file, p);
 			for (struct file_list *l2 = list2; l2; l2 = l2->next) {
 				const char *stripped2 = stripped(l2->file, p);
-				if (!strcmp(stripped1, stripped2)) {
+				if (!strcmp(stripped1, stripped2))
 					return p;
-				}
 			}
 		}
 	}
@@ -832,9 +881,9 @@ output_patch1_only (FILE *p1, FILE *out, int not_reverted)
 
 	fflush (NULL);
 	char *argv[2 + num_diff_opts + 2 + 1];
-	memcpy (argv, (const char *[]) { DIFF, options }, 2 * sizeof (char *));
+	memcpy (argv, ((const char *[]) { DIFF, options }), 2 * sizeof (char *));
 	memcpy (argv + 2, diff_opts, num_diff_opts * sizeof (char *));
-	memcpy (argv + 2 + num_diff_opts, (char *[]) { tmpp1, tmpp2, NULL }, (2 + 1) * sizeof (char *));
+	memcpy (argv + 2 + num_diff_opts, ((char *[]) { tmpp1, tmpp2, NULL }), (2 + 1) * sizeof (char *));
 	in = xpipe (DIFF, &child, "r", argv);
 
 	/* Eat the first line */
@@ -1050,15 +1099,21 @@ trim_context (FILE *f /* positioned at start of @@ line */,
 			printf ("Trim: %lu,%lu\n", strip_pre, strip_post);
 
 		fsetpos (f, &pos);
-		fprintf (out, "@@ -%lu", orig_offset);
-		if (new_orig_count != 1)
-			fprintf (out, ",%lu", new_orig_count);
-		fprintf (out, " +%lu", new_offset);
-		if (new_new_count != 1)
-			fprintf (out, ",%lu", new_new_count);
-		fprintf (out, " @@\n");
+		if (new_orig_count != 1 && new_new_count != 1)
+			print_color (out, LINE_HUNK, "@@ -%lu,%lu +%lu,%lu @@\n",
+				     orig_offset, new_orig_count, new_offset, new_new_count);
+		else if (new_orig_count != 1)
+			print_color (out, LINE_HUNK, "@@ -%lu,%lu +%lu @@\n",
+				     orig_offset, new_orig_count, new_offset);
+		else if (new_new_count != 1)
+			print_color (out, LINE_HUNK, "@@ -%lu +%lu,%lu @@\n",
+				     orig_offset, new_offset, new_new_count);
+		else
+			print_color (out, LINE_HUNK, "@@ -%lu +%lu @@\n",
+				     orig_offset, new_offset);
 
 		while (total_count--) {
+			enum line_type type;
 			ssize_t got = getline (&line, &linelen, f);
 			assert (got > 0);
 
@@ -1070,7 +1125,19 @@ trim_context (FILE *f /* positioned at start of @@ line */,
 			if (total_count < strip_post)
 				continue;
 
-			fwrite (line, (size_t) got, 1, out);
+			switch (line[0]) {
+			case '+':
+				type = LINE_ADDED;
+				break;
+			case '-':
+				type = LINE_REMOVED;
+				break;
+			default:
+				fwrite (line, (size_t) got, 1, out);
+				continue;
+			}
+			print_color (out, type, "%.*s", (int) got - 1, line);
+			fputc ('\n', out);
 		}
 	}
 
@@ -1194,9 +1261,9 @@ output_delta (FILE *p1, FILE *p2, FILE *out)
 	fflush (NULL);
 
 	char *argv[2 + num_diff_opts + 2 + 1];
-	memcpy (argv, (const char *[]) { DIFF, options }, 2 * sizeof (char *));
+	memcpy (argv, ((const char *[]) { DIFF, options }), 2 * sizeof (char *));
 	memcpy (argv + 2, diff_opts, num_diff_opts * sizeof (char *));
-	memcpy (argv + 2 + num_diff_opts, (char *[]) { tmpp1, tmpp2, NULL }, (2 + 1) * sizeof (char *));
+	memcpy (argv + 2 + num_diff_opts, ((char *[]) { tmpp1, tmpp2, NULL }), (2 + 1) * sizeof (char *));
 	in = xpipe (DIFF, &child, "r", argv);
 
 	/* Eat the first line */
@@ -1264,13 +1331,13 @@ output_delta (FILE *p1, FILE *p2, FILE *out)
 				d = *q;
 				*q = '\0';
 			}
-			fprintf (out, DIFF " %s %s %s\n", options, oldname + 4,
-				 newname + 4);
+			print_color (out, LINE_HEADER, DIFF " %s %s %s\n", options, oldname + 4,
+				     newname + 4);
 			if (p) *p = c;
 			if (q) *q = d;
 		}
-		fprintf (out, "--- %s\n", oldname + 4);
-		fprintf (out, "+++ %s\n", newname + 4);
+		print_color (out, LINE_FILE, "--- %s\n", oldname + 4);
+		print_color (out, LINE_FILE, "+++ %s\n", newname + 4);
 		rewind (tmpdiff);
 		trim_context (tmpdiff, file.unline, out);
 		fclose (tmpdiff);
@@ -1376,29 +1443,36 @@ index_patch_generic (FILE *patch_file, struct file_list **file_list, int need_sk
 
 		/* For patch2, we need to handle the @@ line and skip content */
 		if (need_skip_content) {
-			if (getline (&line, &linelen, patch_file) == -1) {
+			int found = 0;
+
+			while (!found &&
+			       getline (&line, &linelen, patch_file) > 0) {
+				if (strncmp (line, "@@ ", 3))
+					continue;
+
+				p = strchr (line + 3, '+');
+				if (!p)
+					continue;
+
+				p = strchr (p, ',');
+				if (p) {
+					/* Like '@@ -1,3 +1,3 @@' */
+					p++;
+					skip = strtoul (p, &end, 10);
+					if (p == end)
+						continue;
+				} else
+					/* Like '@@ -1 +1 @@' */
+					skip = 1;
+				found = 1;
+			}
+
+			if (!found) {
 				free (names[0]);
 				free (names[1]);
 				break;
 			}
 
-			if (strncmp (line, "@@ ", 3))
-				goto try_next;
-
-			p = strchr (line + 3, '+');
-			if (!p)
-				goto try_next;
-			p = strchr (p, ',');
-			if (p) {
-				/* Like '@@ -1,3 +1,3 @@' */
-				p++;
-				skip = strtoul (p, &end, 10);
-				if (p == end)
-					goto try_next;
-			} else
-				/* Like '@@ -1 +1 @@' */
-				skip = 1;
-
 			add_to_list (file_list, best_name (2, names), pos);
 
 			while (skip--) {
@@ -1414,7 +1488,6 @@ index_patch_generic (FILE *patch_file, struct file_list **file_list, int need_sk
 			add_to_list (file_list, best_name (2, names), pos);
 		}
 
-	try_next:
 		free (names[0]);
 		free (names[1]);
 	}
@@ -1422,10 +1495,26 @@ index_patch_generic (FILE *patch_file, struct file_list **file_list, int need_sk
 	if (line)
 		free (line);
 
+	/* Return success if the file is empty, has indexed files, or
+	 * has no --- lines at all (merge commits, mode-only changes,
+	 * notes-only commits, binary-only patches, etc.). */
 	if (file_is_empty || *file_list)
 		return 0;
-	else
-		return 1;
+
+	/* Check if the patch has any file-header --- lines (as
+	 * opposed to commit-message separators or b4 tracking).
+	 * If not, it simply has no diffable content. */
+	line = NULL;
+	linelen = 0;
+	rewind (patch_file);
+	while (getline (&line, &linelen, patch_file) != -1)
+		if (!strncmp (line, "--- a/", 6) ||
+		    !strncmp (line, "--- /dev/null", 13)) {
+			free (line);
+			return 1;
+		}
+	free (line);
+	return 0;
 }
 
 static int
@@ -1613,9 +1702,9 @@ take_diff (const char *f1, const char *f2, char *headers[2],
 		sprintf (options, "-U%d", max_context);
 
 	char *argv[2 + num_diff_opts + 2 + 1];
-	memcpy (argv, (const char *[]) { DIFF, options }, 2 * sizeof (char *));
+	memcpy (argv, ((const char *[]) { DIFF, options }), 2 * sizeof (char *));
 	memcpy (argv + 2, diff_opts, num_diff_opts * sizeof (char *));
-	memcpy (argv + 2 + num_diff_opts, (const char *[]) { f1, f2, NULL }, (2 + 1) * sizeof (char *));
+	memcpy (argv + 2 + num_diff_opts, ((const char *[]) { f1, f2, NULL }), (2 + 1) * sizeof (char *));
 	if (debug) {
 		fputs ("+", stdout);
 		for (int i = 0; argv[i]; i++) {
@@ -2047,11 +2136,15 @@ interdiff (FILE *p1, FILE *p2, const char *patch1, const char *patch2)
 
 		names[0] = filename_from_header (line + 4);
 
-		if (getline (&line, &linelen, p1) == -1)
+		if (getline (&line, &linelen, p1) == -1) {
+			free (names[0]);
 			break;
+		}
 
-		if (strncmp (line, "+++ ", 4))
+		if (strncmp (line, "+++ ", 4)) {
+			free (names[0]);
 			continue;
+		}
 
 		names[1] = filename_from_header (line + 4);
 
@@ -2084,8 +2177,21 @@ interdiff (FILE *p1, FILE *p2, const char *patch1, const char *patch2)
                 free (p);
 	}
 
-	if (!file_is_empty && !patch_found)
-		no_patch (patch1);
+	if (!file_is_empty && !patch_found) {
+		/* Don't warn for patches with no file-header --- lines
+		 * (merge commits, mode-only, binary-only, etc.). */
+		int has_diff_content = 0;
+
+		rewind (p1);
+		while (getline (&line, &linelen, p1) != -1)
+			if (!strncmp (line, "--- a/", 6) ||
+			    !strncmp (line, "--- /dev/null", 13)) {
+				has_diff_content = 1;
+				break;
+			}
+		if (has_diff_content)
+			no_patch (patch1);
+	}
 
 	copy_residue (p2, mode == mode_flip ? flip1 : stdout);
 
@@ -2096,12 +2202,10 @@ interdiff (FILE *p1, FILE *p2, const char *patch1, const char *patch2)
 
 		if (flipdiff_inplace) {
 			/* Use atomic in-place writing for safety */
-			if (write_file_inplace(patch2, flip1) != 0) {
+			if (write_file_inplace(patch2, flip1) != 0)
 				error (EXIT_FAILURE, errno, "failed to write %s", patch2);
-			}
-			if (write_file_inplace(patch1, flip2) != 0) {
+			if (write_file_inplace(patch1, flip2) != 0)
 				error (EXIT_FAILURE, errno, "failed to write %s", patch1);
-			}
 		} else {
 			copy (flip1, stdout);
 			puts ("\n=== 8< === cut here === 8< ===\n");
@@ -2280,12 +2384,12 @@ main (int argc, char *argv[])
 			const char *color_mode = optarg ? optarg : "auto";
 
 			/* Handle auto mode: check if stdout is a terminal */
-			if (strcmp(color_mode, "auto") == 0) {
+			if (strcmp(color_mode, "auto") == 0)
 				color_mode = isatty(STDOUT_FILENO) ? "always" : "never";
-			}
 
-			if (asprintf (diff_opts + num_diff_opts++, "--color=%s", color_mode) < 0)
-				error (EXIT_FAILURE, errno, "Memory allocation failed");
+			/* Set our internal color flag instead of passing to diff */
+			use_colors = (strcmp(color_mode, "always") == 0);
+			color_option_specified = 1;
 			break;
 		}
 		case 1000 + 'I':
@@ -2317,18 +2421,9 @@ main (int argc, char *argv[])
 		error (EXIT_FAILURE, 0,
 		       "-z and --in-place are mutually exclusive.");
 
-	/* Add default color=always if no color option was specified and we're in a terminal */
-	int has_color_option = 0;
-	for (int i = 0; i < num_diff_opts; i++) {
-		if (strncmp(diff_opts[i], "--color", 7) == 0) {
-			has_color_option = 1;
-			break;
-		}
-	}
-	if (!has_color_option && isatty(STDOUT_FILENO)) {
-		if (asprintf (diff_opts + num_diff_opts++, "--color=always") < 0)
-			error (EXIT_FAILURE, errno, "Memory allocation failed");
-	}
+	/* Set default color behavior if no color option was specified */
+	if (!color_option_specified && isatty(STDOUT_FILENO))
+		use_colors = 1;
 
 	if (optind + 2 != argc)
 		syntax (1);
